#include <algorithm>
#include <complex>
#include <cstdlib>
#include <ctime>
#include <functional>
#include <iostream>
#include <iomanip>
#include <iterator>

/**
* Auxiliary: sine wave generator to initiate input test data samples
*/
template <typename T, size_t period = 1>
class sin_gen
{
   T i;
public:
   sin_gen(T const &_i = T() - 1):i(_i)  {  ;  }
   std::complex<T> operator ()()
   {
      using std::sin;
      return ++i, sin(_2PI_period * i);
   }
   std::complex<T> operator ()(T const ii)
   {
      using std::sin;
      return ++i, sin(_2PI_period * ii);
   }
   static long double const PI;
   static long double const _2PI_period;
};

template <typename T, size_t period>
long double const sin_gen<T, period>::PI = 3.14159265358979323846L;

template <typename T, size_t period>
long double const sin_gen<T, period>::_2PI_period
                                            = 2 * sin_gen<T, period>::PI/period;

inline
double my_clock()
{
   return std::clock()/static_cast<double>(CLOCKS_PER_SEC);
}

/** Bit reversal routine
*   Used for reordering of input data for FFT in Danielson-Lanczos algorithm).
*   Note: CT loop unrolling is applied since bits is of resonable size.
*   It is not always the case that loop unrolling will improve the algorithm's
*   performance. Actually, it doesn't for Athlon XP for bits > 4, but succeeds
*   for bcc32 code and at UltraSPARC III plaform.
*/
namespace CT_RT
{
   template <typename T, size_t bits>
   struct bit_reverse
   {
      static inline T reverse(T const i)
      {
         T t = (1 & i) << bits - 1;
         return t | bit_reverse<T, bits - 1>::reverse(i >> 1);
      }
   };

   template <typename T>
   struct bit_reverse<T, 0>
   {
      static inline T reverse(T const i)  {  return 0;  }
   };
}

namespace RT
{
   template <typename T, size_t bits>
   struct bit_reverse
   {
      static size_t reverse(T n, size_t b = bits);
   };

   template <typename T, size_t bits>
   size_t bit_reverse<T, bits>::reverse(T n, size_t b)
   {
      T t = T();
      while(b--)
      {
         t <<= 1;
         t  |= 1 & n;
         n >>= 1;
      }
      return t;
   }
}

int main(int, char * const[])
{
   using namespace std;
   typedef complex<double> cplx;

   using RT::bit_reverse;

   // test data preparation
   size_t const bits = 22;
   size_t const n    = 1 << bits;

   cplx *samples = new cplx[n];

   generate(samples, samples + n,
            sin_gen<double, n>());
//   transform(samples, samples + n,
//             samples,
//             bind2nd(plus<cplx>(), cplx(0.5)));

   // original samples
   if(bits < 5)
   {
      cout << setprecision(4) << fixed;
      copy(samples, samples + n,
           ostream_iterator<cplx, char>(cout, "\t"));
      cout << endl;
   }
   double start = my_clock();

   /** FFT transform (local)
   *                 (Danielson - Lanczos algorithm; cf. NRC: s. 12.2)
   *                 http://www.library.cornell.edu/nr/bookcpdf/c12-2.pdf
   *                 (Arndt implementation; cf A4P, s. 1.3)
   *                 http://www.jjj.de/fxt/
   * 1. Bit reversal swapping - 1-point transform
   */
   size_t const bits2 = bits/2 + bits%2;
   size_t const br_n  = n - (1 << bits2);
      /** slk: 'br_n' is the maximum index to be swapped. Therefore there
      *        is no need to continue reversing for the remainder. The actual
      *        gain decreases with growing 'bits', and for instance, for
      *        bits = 10 (1024 samples) it is only c.a. 3% (however - for free).
      */

   for(size_t i = 0; i < br_n; ++i)
   {
      size_t r = bit_reverse<size_t, bits>::reverse(i);
      if(i < r) swap(samples[i], samples[r]);
   }
   /**
   * 2. Actual FFT transform
   */
   for(size_t i = 1; i < bits + 1; ++i)   // log_2(n) iterations
   {
      // start from 2-point transform and continue with
      // 4, 8, 16,..., (2^bits)-point ones
      size_t j  = 1 << i;
      size_t j2 = j >> 1;

      // e_phi is used to iterative computation of powers of w
      cplx const phi(0.0, 2.0 * sin_gen<double>::PI/j);
      cplx const e_phi(exp(phi));

      for(size_t k = 0; k < n; k += j)    // n/j iterations
      {
         cplx w(1.0);
         for(size_t l = 0; l < j2; ++l)   // j/2 iterations (Butterfly?)
         {
            // computing indices
            size_t const kl   = k  + l,
                         klj2 = kl + j2;
            // computing temporaries (variable 'u' only improves readability)
            cplx const u = samples[kl],
                       v = samples[klj2] * w;
            // replacing the (j-1)-point transform with the j-point one
            samples[klj2] = u - v;
            samples[kl]  += v;
            // computing the next power of 'w' (nlog_2(n) times)
            w *= e_phi;
         }
      }
   }
   /** End of FFT transform.
   *   Observe: the algorithm works 'in situ'
   */
   cout << my_clock() - start << endl;
   start = my_clock();
   /** Inverse FFT (non-local)
   *               (Arndt implementation; cf A4P, s. 1.3) (*)
   *               http://www.jjj.de/fxt/
   * 1. Bit reversal swapping - 1-point transform
   */
   for(size_t i = 0; i < br_n; ++i)
   {
      size_t r = bit_reverse<size_t, bits>::reverse(i);
      if(i < r) swap(samples[i], samples[r]);
   }
   /**
   * 2. Inverse FFT transform
   */
   for(size_t i = 1; i < bits + 1; ++i)   // log_2(n) iterations
   {
      // start from 2-point transform and continue with
      // 4, 8, 16,..., (2^bits)-point ones
      size_t j  = 1 << i;
      size_t j2 = j >> 1;

      // e_phi is used to iterative computation of powers of w
      cplx const phi(0.0, -2.0 * sin_gen<double>::PI/j);
      cplx const e_phi(exp(phi));
      cplx       w(1.0);

      for(size_t l = 0; l < j2; ++l)      // j/2 iterations (Butterfly?)
      {
         for(size_t k = 0; k < n; k += j) // n/j iterations
         {
            // computing indices
            size_t const kl   = k  + l,
                         klj2 = kl + j2;
            // computing temporaries
            cplx const u = samples[kl],
                       v = samples[klj2] * w;
            // replacing the (j-1)-point transform with the j-point one
            samples[klj2] = u - v;
            samples[kl]   = u + v;
         }
         // computing the next power of w (n times)
         w *= e_phi;
      }
   }
   /** End of inverse FFT transform.
   *   Observe: the algorithm works 'in situ'
   */
   cout << my_clock() - start << endl;
   // original samples?
   if(bits < 5)
   {
      copy(samples, samples + n,
           ostream_iterator<cplx, char>(cout, "\n"));
//      transform(samples, samples + n,
//                ostream_iterator<cplx, char>(cout, "\n"),
//                bind2nd(divides<cplx>(), cplx(n)));
   }
   delete []samples;
   return 0;
}

//      /**
//      * 2. FFT transform with unrolling
//      */
//      for(size_t i = 1; i < bits + 1; ++i)   // log_2(n) iterations
//      {
//         // start from 2-point transform and continue with
//         // 4, 8, 16,..., (2^bits)-point ones
//         size_t j  = 1 << i;
//         size_t j2 = j >> 1;
//
//         // e_phi is used to iterative computation of powers of w
//         cplx const phi(0.0, 2.0 * sin_gen<double>::PI/j);
//         cplx const e_phi(exp(phi));
//
//         for(size_t k = 0; k < n; k += j)    // n/j iterations
//         {
//            cplx w(1.0);
//            for(size_t l = 0; l < j2; l += 2)   // j/4 iterations (Butterfly?)
//                                                // with primitive loop unrolling
//            {
//               // computing indices
//               size_t const kl   = k  + l,
//                            klj2 = kl + j2;
//               // computing temporaries (variable u only improves readability)
//               cplx const u  = samples[kl],
//                          u2 = samples[kl + 1],
//                          v  = samples[klj2] * w,
//                          v2 = samples[klj2 + 1] * (w *= e_phi, w);
//               // replacing the (j-1)-point transform with the j-point one
//               samples[klj2] = u - v;
//               samples[kl]  += v;
//
//               samples[klj2 + 1] = u2 - v2;
//               samples[kl + 1]  += v2;
//               // computing the next power of w (nlog_2(n) times)
//               w *= e_phi;
//            }
//         }
//      }
