//---------------------------------------------------------------------------

#pragma hdrstop
#include <iostream>
#include <math>

//---------------------------------------------------------------------------

/**
   The nth Fibonacci number can be computed directly as ([1, 2, 3]):

      $F_{n}=\left[ \frac{1}{\sqrt{5}}
             \left( \frac{1+\sqrt{5}}{2}\right) ^{n}\right]
            =\left[ \frac{\phi ^{n}}{\sqrt{5}}\right]$

   where $\phi$ is the golden ratio:

      $\frac{1}{1-\phi }=\phi $

  But the really jaw-dropping are the following equality:

   $\phi =\sqrt{1+\sqrt{1+\sqrt{1+\sqrt{1+\sqrt{1+\cdots }}}}}$

  or

   $\phi = 1+\frac{1}{1+\frac{1}{1+\frac{1}{1+\frac{1}{1+\cdots }}}}$.
  */

template <typename I, typename F>
class _fibonacci
{
public:
   static const F golden_ratio;
   static const F sqrt_5;
   I n;
public:
   _fibonacci<I, F>(I _n):n(_n)   {  }
   I operator ()() const;
};

template <typename I, typename F>
inline
I _fibonacci<I, F>::operator ()() const
{
   return std::floor(0.5 + std::pow(golden_ratio, n)/sqrt_5);
}

template <typename I, typename F>
const F _fibonacci<I, F>::sqrt_5 = static_cast<F>(sqrt(5.0));

template <typename I, typename F>
const F _fibonacci<I, F>::golden_ratio = static_cast<F>((1.0 + sqrt(5.0))/2.0);

// caling facility
template <typename I, typename F>
inline
I fibonacci(I n)
{
   return _fibonacci<I, F>::_fibonacci(n)();
}

/**
   Templates that can be used to generate Fibonacci numbers recursively:
   1. during compilation
  */

template <typename I, int n>
struct __fibonacci
{
   enum { _n = __fibonacci<I, n - 1>::_n + __fibonacci<I, n - 2>::_n   };
};

template <typename I>
struct __fibonacci<I, 2>
{
   enum { _n = 1   };
};

template <typename I>
struct __fibonacci<I, 1>
{
   enum { _n = 1   };
};

// caling facility
template <typename I, int n>
inline
I fibonacci()
{
   return __fibonacci<I, n>::_n;
}

/**
   2. during compilation (or run-time) - it depends on the compiler how it
      reacts on 'inline' keywords... (KSB pointed it out to me.)
  */

template <int n>
inline
int fibonacci()
{
   return fibonacci<n - 1>() + fibonacci<n - 2>();
}

template <>
inline
int fibonacci<2>()
{
   return 1;
}

template <>
inline
int fibonacci<1>()
{
   return 1;
}

/**
   or:
   3. during run-time rather (or still during compilation, when the argument is
      a constant and compiler is smart enough!)
  */
template <typename I>
inline
I fibonacci(I const n)
{
   return n > 2 ? fibonacci(n - 1) + fibonacci(n - 2)
                : 1;
}


/**
Bibliography:
-----
 [1] Leonardo Bonacci (Fibonacci), Liber abaci, Piza, 1202.

 [2] D. Wells, The Penguin Dictionary of Curious and Interesting Numbers.
     Middlesex, England: Penguin Books, pp. 61-67, 1986.

 [3] Mathematica's WWW page: http://mathworld.wolfram.com/FibonacciNumber.html
  */

int main(int, char *[])
{

   using namespace std;

   // recursion performed by compiler
   int i = fibonacci<10>();
   cout << i << endl;

   // recursion performed in run-time
   // (but maybe smart compiler could cope with it?)
   i = fibonacci(10);
   cout << i << endl;

   // non-recursive algorithm performed in run-time
   // (but maybe smart compiler would dare again?)
   i = fibonacci<int, double>(11);
   cout << i << endl;

   // ditto?
   i = fibonacci<int, double>(i);
   cout << i << endl;

   // only compile-time recursion
   i = fibonacci<long, 10>();
   cout << i << endl;

   return 0;
}
//---------------------------------------------------------------------------

