#include <functional>
#include <iostream>
#include <cmath>

template <typename T, template <typename A, typename R = A> class F = std::pointer_to_unary_function>
class piecewise_constant: public std::unary_function<T, T>
{
   typedef T argument_type;
   typedef T result_type;

   F<T, T> const &f;
   T       const grid, dirg;
public:
   piecewise_constant(F<T, T> const &_f, T g): f(_f), grid(g), dirg(1.0/g)
   {  ;  }
   T operator ()(T x) const
   {
      using std::floor;

      T const x_0 = floor(x * dirg + 0.5) * grid;
      return f(x_0);
   }
};

template <typename T, template <typename A, typename R = A> class F = std::pointer_to_unary_function>
class piecewise_linear: public std::unary_function<T, T>
{
   typedef T argument_type;
   typedef T result_type;

   F<T, T> const &f;
   T       const dirg, grid;
public:
   piecewise_linear(F<T, T> const &_f, T const g): f(_f), grid(g), dirg(1.0/g)
   {  ;  }
   T operator ()(T x) const
   {
      using std::floor;

      T const x_0 = floor(x * dirg) * grid,
              x_1 = x_0 + grid;
      T const y_0 = f(x_0),
              y_1 = f(x_1);
      // spline-style version:
      // return dirg * ((x_1 - x) * y_0 + (x - x_0) * y_1);
      // slk's version:
      return y_0 + dirg * (x - x_0) * (y_1 - y_0);
   }
};

template <typename function>
class piecewiser: public std::unary_function<typename function::argument_type,
                                             typename function::result_type>
{
   typedef typename function::argument_type argument_type;
   typedef typename function::result_type   result_type;

   function       f;
   argument_type  const dirg, grid;
public:
   piecewiser(function const &_f, argument_type const g): f(_f), grid(g), dirg(1.0/g)
   {  ;  }
   result_type operator ()(argument_type x) const
   {
      using std::floor;

      argument_type const x_0 = floor(x * dirg) * grid,
                          x_1 = x_0 + grid;
      result_type const   y_0 = f(x_0),
                          y_1 = f(x_1);
      // spline-style version:
      // return dirg * ((x_1 - x) * y_0 + (x - x_0) * y_1);
      // slk's version:
      return y_0 + dirg * (x - x_0) * (y_1 - y_0);
   }
};

template <typename function>
inline
piecewiser<function> piecewise(function const &f, typename function::argument_type dirg)
{
   return piecewiser<function>(f, dirg);
}

template <typename A, typename R = A>
struct mfo: std::unary_function<A, R>
{
   typedef A argument_type;
   typedef R result_type;

   mfo(A = A())  { }
   result_type operator ()(argument_type x) const
   {
      using  std::sin;
      return sin(x);
   }
};

template <typename A>
inline A mf(A x)
{
   return mfo<A, A>()(x);
}

int main(int, char * const[])
{
   using std::cout;
   using std::endl;
   using std::ptr_fun;

   double dirg = 0.125;
   /** if you adopt existing function...
   *   or if use a function object.
   */
   piecewise_constant<double>    pc(ptr_fun(mf<double>), dirg);
   mfo<double>                   mfod;
   piecewise_linear<double, mfo> pl(mfod, dirg);

   for(mfo<double>::argument_type x = 0; x < 6.29; x += 0.01)
   {
      cout << x      << "\t"
           << ml(x)  << "\t"
           << pl2(x) << "\t"
           << pl(x)  << "\t"
           << pl2(x) << endl;
   }
   return 0;
}
