//---------------------------------------------------------------------------

#pragma hdrstop

//---------------------------------------------------------------------------

#include <functional>
#include <iostream>
#include <iterators>
#include <vector>

////////////////////////////////////////////////////////////////////////////////
////////////////  Experimental zone, X-Files protected   ///////////////////////
////////////////////////////////////////////////////////////////////////////////

namespace slk
{

template <class Operation>
class binder2nd: public std::unary_function<typename Operation::first_argument_type,
                                            typename Operation::result_type>
{
   // express types of 'binder' in terms of 'Operation' ones
   typedef typename std::unary_function<typename Operation::first_argument_type,
                                        typename Operation::result_type>::argument_type argument_type;

   typedef typename std::unary_function<typename Operation::first_argument_type,
                                        typename Operation::result_type>::result_type result_type;

   typedef typename Operation::second_argument_type value_type;
protected:
   Operation   op;
   value_type  value;
public:
   // The constructor (do almost nothing;)
   // Not really: initiates object's fields.
   binder2nd(const Operation &, const value_type &);

   result_type operator() (const argument_type &) const;
};

template <class Operation>
binder2nd<Operation>::binder2nd(const Operation& o, const value_type &v)
                     :op(o), value(v)
{
}

template <class Operation>
inline binder2nd<Operation>::result_type
binder2nd<Operation>::operator() (const binder2nd<Operation>::argument_type &x) const
{
   return op(x, value);
}

// The function determines proper template types, hence we can avoid the
// tedious job of resolving them manually.
template <class Operation, class T>
inline binder2nd<Operation> bind2nd(const Operation& op, const T& x)
{
   typedef typename Operation::second_argument_type argument_type;
   return binder2nd<Operation>(op, argument_type(x));
}


// namespace
};

////////////////////////////////////////////////////////////////////////////////
////////////////  Experimental zone, X-Files protected   ///////////////////////
////////////////////////////////////////////////////////////////////////////////

template <typename T>
class counter
{
   static T object_counter;
   T object_number;
public:
   // standard and copy constructors...
   counter(T const & t = object_counter):object_number(t) {  ++object_counter; }
   counter(counter const &)               {  object_number = object_counter++; }

   // ... and assignment operator
   counter &operator =(counter const &);

   virtual ~counter()           {  --object_counter; }

   // A method used by collection of pointers
   virtual counter *increase()  {  return ++object_number,  this; }
   // A method used by collection of objects
   virtual counter &decrease()  {  return --object_number, *this; }

   virtual void print() const   {   cout << "I'm: " << object_number << endl;  }

   // Methods used to create/destroy objects in collection of pointers
   static counter *create();
   static counter *destroy(counter *);
};

// Be carefull, if one use specializations...
template <typename T>
T counter<T>::object_counter = 0;

template <typename T>
inline
counter<T> &counter<T>::operator =(counter<T> const &)
{
   return object_number = object_counter++, *this;
}

template <typename T>
inline
counter<T> *counter<T>::create()
{
   return new counter<T>();
}

/** The following returns pointer since 'transform' algorithm needs the return
     value. (cf. B. Stroustrup, The C++ Programming Language, 3rd ed.)
  */
template <typename T>
inline
counter<T>* counter<T>::destroy(counter<T> *p)
{
   // Zeroing the pointer allows to 'safely delete' the pointed object and is
   // a by-product of the necessity of returning a pointer mentioned above.
   return delete p, reinterpret_cast<counter<T>*>(0);
}

/** More tricky (and more universal, moreover) solution is to create an output
    iterator to empty collection (a kind of 'devnull') - according to the
    BS's idea.
    The following do nothing but matches the standard library programming style.

    copy(vci.begin(), vci.end(),
         devnull_iterator<counter<int>, char>());
  */
using namespace std;

template <class T, class charT = char, class traits = char_traits<charT> >
class devnull_iterator : public iterator<output_iterator_tag, T, ptrdiff_t, T*, T&>
{
public:
   // This makes the iterator the output one.
   devnull_iterator<T, charT, traits> &operator =(const T&)   { return *this; }

   // They make the iterator itself.
   devnull_iterator<T, charT, traits> &operator  *()          { return *this; }
   devnull_iterator<T, charT, traits> &operator ++()          { return *this; }
   devnull_iterator<T, charT, traits> &operator ++(int)       { return *this; }
};

/** A generic stuff converting pointers to objects
  */

// A function...
template <typename T>
inline
T pointer_to_value_pf(T const *p)
{
   return *p;
}

// ... and function object.
template <typename T>
class pointer_to_value_of : public unary_function<T*, T>
{
public:
   T operator ()(T const *t)
   {
      return *t;
   }
};


/** Fortunatelly, here this nasty/vicious template specialization is valid for
    the current file. However one can easily imagine the situation in which one
    specialize a template which is then used be exported methods...
  */
template <>
int counter<int>::object_counter = 12;


int main(int, char *[])
{
   int it[10] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};
   remove_if(it, it + 10, slk::bind2nd(equal_to<int>(), 3));
   copy(it, it + 10, ostream_iterator<int, char>(cout, " "));


   vector<counter<int> *> vpci(12);

   // create a dozen of pointers to objects of 'counter<int>'
   generate(vpci.begin(), vpci.end(),
            counter<int>::create);

   /** Use 'transform' algorithm to emphasize that the sequence is changed.
       Observe that the following
       for_each(vci.begin(), vci.end(),
            mem_fun_ref(counter<int>::decrease));
       is also possible, however, in Standard C++ Library reference 'for_each'
       is classified as non-modifying algorithm. 
     */
   transform(vpci.begin(), vpci.end(),
             vpci.begin(),
             mem_fun(counter<int>::increase));

   // Use 'for_each' algorithm to point out that the sequence remains unchanged
   for_each(vpci.begin(), vpci.end(),
            mem_fun(counter<int>::print));

   vector<counter<int> > vci(12);
   for_each(vci.begin(), vci.end(),
            mem_fun_ref(counter<int>::print));

   // copy the collection of 'counter<int> *' to the collection
   // of ordinary 'counter<int>' type:

   // Using function template...
   transform(vpci.begin(), vpci.end(),
             vci.begin(),
             pointer_to_value_pf<counter<int> >);
   // ... and using function object
   transform(vpci.begin(), vpci.end(),
             vci.begin(),
             pointer_to_value_of<counter<int> >());


   transform(vci.begin(), vci.end(),
             vci.begin(),
             mem_fun_ref(counter<int>::decrease));
   for_each(vci.begin(), vci.end(),
            mem_fun_ref(counter<int>::print));

   //destroy all 'counter<int>' objects within 'vpci'
   transform(vpci.begin(), vpci.end(),
             vpci.begin(),
             counter<int>::destroy);

   /** It is safe to destroy counters twice (or more) using
      'counter<int>::destroy' because 'vpci' contains pointers set to 0 after
      the first transform call.
     */
   transform(vpci.begin(), vpci.end(),
             devnull_iterator<counter<int> *, char>(),
             counter<int>::destroy);
   return 0;
}
//---------------------------------------------------------------------------

