//---------------------------------------------------------------------------

#pragma hdrstop
#include <iostream>

//---------------------------------------------------------------------------


/**
   AUTHOR:  P. liwinski
   PROJECT: Implementation of n-tuple
   VERSION: 0.1
   The actual aim of the following code is to present a technique to implement
   inline (i.e. without unnecesarry temporary object) ternary (quadruple,
   quintuple, sextuple etc) operators.

   EXAMPLE

      Consider an operation:

         x = a * b + c;                                                      (1)

      where 'a', 'b' and 'c' are vectors/matrices or any non-scalar.

      Typically, first the product is computed and the temporary object
      'ab' is constructed to store it. Then the sum of 'ab' and 'c' is
      calculated.

      Hence, any C geek/nerd must say: "That's all so hyped C++'s overloading
      operators... for nothing! I will simply write a function
      'mul_and_add(a, b, c)'! and (s)he is undoubtedly right!

      So... let's take that function from the C fan and...

   SOLUTION

      When tuning the calculation (1), removing temporaries is only the half
      of the way. Obviosly, C's function still remains faster since it performs
      computation in one loop, while we do it in two: one for product and one
      for computing the sum.

      Thus:
      1. Create the operator '*' which produce an object (class '..._product',
         say, which does nothing except storing references to 'a' and 'b'.

      2. Implement the operator '+' for the class 'product' and the type of 'c'.
         This operator has at its disposal reference to 'a', 'b' and 'c'. Hence,
         it can call our C function.

      3. If both operators are inline then writing (1) becomes equivalent to
         'mul_and_add(a, b, c)'!
  */


// necessary forward declarations
template <typename T>
class matrix;

template <typename T>
T mull_and_add(matrix<T> const &, matrix<T> const &, matrix<T> const &);

template <typename T>
class matrix_product;

template <typename T>
T operator +(matrix_product<T> const &, matrix<T> const &);


// ...given the existing type, a 'matrix', say, create the operator '=*+'

template <typename T>
class matrix
{
   // data
   T value;                                       
public:
   matrix(T val = T(0)):value(val) {  }

   // interface
   friend T mull_and_add<>(matrix<T> const &,
                           matrix<T> const &,
                           matrix<T> const &);
};

template <typename T>
inline
T mull_and_add(matrix<T> const &a, matrix<T> const &b, matrix<T> const &c)
{
   //compute product and sum in one loop
   return a.value * b.value + c.value;
}


//  matrix product
template <typename T>
class matrix_product
{
   matrix<T> const &a, &b;
public:
   matrix_product(matrix<T> const &_a, matrix<T> const &_b):a(_a), b(_b)  {  }

   friend T operator +<>(matrix_product<T> const &, matrix<T> const &);
};

template <typename T>
inline
matrix_product<T> operator *(const matrix<T> &a, const matrix<T> &b)
{
   return matrix_product<T>(a, b);
}

template <typename T>
inline
T operator +(matrix_product<T> const &ab, matrix<T> const &c)
{
   return mull_and_add(ab.a, ab.b, c);
}

#pragma argsused
int main(int, char *[])
{
   using namespace std;

   matrix<double> a(2),
                  b(4),
                  c(2);
   double mac = a * b + c;

   double _a(2),
          _b(4),
          _c(2);

   double _mac = _a * _b + _c;

   cout << mac  << endl
        << _mac << endl;
   return 0;
}
//---------------------------------------------------------------------------
