//---------------------------------------------------------------------------
#pragma hdrstop
#include <algorithm>
#include <iostream>
#include <functional>
#include <limits>
#include <valarray>
//---------------------------------------------------------------------------
using namespace std;

template <typename T>
class matrix
{
   valarray<T> &v;
   slice const s;
public:
   matrix(valarray<T> &_v, const size_t rows, const size_t columns):
          v(_v),
          s(0, rows, columns)
   {
   }
   T &operator ()(size_t i, size_t j)
   {
      return v[s.start() + s.stride() * i + j];
   }
   T operator ()(size_t i, size_t j) const
   {
      return this->operator ()(i, j);
   }
   T *begin() {  return &v[0];        }
   T *end()   {  return &v[v.size()]; }
};

template <class Container>
class x_2_alpha_insert_iterator:
      public iterator<output_iterator_tag, void, void, void, void>
{
protected:
   Container &container;
public:
   typedef Container container_type;

   explicit x_2_alpha_insert_iterator(Container& x): container(x) {}
   x_2_alpha_insert_iterator<Container>& operator =(Container::const_reference v)
   {
      //
      return *this;
   }
   x_2_alpha_insert_iterator<Container>& operator*  ()    { return *this; }
   x_2_alpha_insert_iterator<Container>& operator++ ()    { return *this; }
   x_2_alpha_insert_iterator<Container>  operator++ (int) { return *this; }
};

template <class Container>
inline x_2_alpha_insert_iterator<Container> x_2_alpha_inserter (Container& x)
{
   return x_2_alpha_insert_iterator<Container>(x);
}


template <typename T>
class random_gen
{
public:
   random_gen(T seed = T(0))  {  srand(seed ? seed : time(0)); }
   T operator ()() const      {  return rand()/static_cast<T>(RAND_MAX); }
};


#pragma argsused
int main(int argc, char* argv[])
{
   size_t        size(6);
   valarray<int> va(size * size);
   matrix<int>   mi(va, size , size);

   generate(mi.begin(), mi.end(),
            random_gen<int>());

   copy(mi.begin(), mi.end(),
        ostream_iterator<int, char>(cout, "\n"));

   return 0;
}
//---------------------------------------------------------------------------
/**
   // copy even values
   slice_array<int> &sb = //va[slice(0, va.size()/2, 2)];
                        // First half:
                        //  slice(0, va.size()/2, 1);
                        // second half:
                        //  slice(5, va.size()/2, 1);
                        // first row of [3][4] matrix:
                        //  va[slice(1, 3, 4)];
                        // first column of [3][4] matrix:
                          va[slice(0, 4, 1)];
  */


/**
   ASSUMPTIONS:
   1. {x} collects random doubles and x_min, x_max are unknown
   2. A scale K is selected according to the number N of measurement data

   ALGORITHM:
   1. Find x_max and x_min;                        complexity: O(N)
   2. Compute coefficients alpha at the scale K;   complexity: O(p*N)
   3. Compute the rest of coefficients;            complexity: O(N)
   -------------------------------------------------------------------
      TOTAL:                                                   O(p*N).
  */

   // 1. Find x_max and x_min
   const double x_min = *min_element(x.begin(), x.end()),
                x_max = *max_element(x.begin(), x.end());

   const int K = 1.0/(2*gamma + 1)*log_2(2*gamma*N);

   const int n_min = floor((1 << K)*x_min - s2),
             n_max = floor((1 << K)*x_max - s1);

   // 2. Compute coefficients alpha[] at the scale K
   scaling_function  phi<double>(p, K);
   array<double>     alpha(n_min, n_max, 0.0);


   transform(x.begin(), x.end()),   // source
             inserter<...>(alpha),  // destimation
             mul_y_phi(phi));

/**
   INTRODUCTION:

   What does the 'transform' algorithm actually do?

   It takes arguments from input sequence(s), operates on it/them, and creates
   the result to put it into the output one.

   Hence the types of the input(s) and the output have only to match to the
   appropriate arguments and result types; see declarations:
...*/

   template <typename Arg1, typename Arg2, typename Result>
   struct binary_function
   {
      typedef Arg1 first_argument_type;
      typedef Arg2 second_argument_type;
      typedef Result result_type;
   };

   template <typename T1, typename T2, typename R>
   struct do_sth : public binary_function<T1, T2, R>
   {
      typedef typename binary_function<T1, T2, R>::second_argument_type second_argument_type;
      typedef typename binary_function<T1, T2, R>::first_argument_type first_argument_type;
      typedef typename binary_function<T1, T2, R>::result_type result_type;

      R operator() (const T1& x, const T2& y) const { return x sth y; }
   };

/**...
   It makes 'transform' really generic and (potentially) powerful!

   --
   EXAMPLE:

   Consider the problem of computing initial coefficients alpha at the scale K
   given the data set {x_k, y_k}, k = 1,...,N, the scale K and p-th scaling
   function.

   Input seqence is ready to use. The clue is the iterator for the output one.
   It should take

   1. The current pair of measurements (from input sequence indirectly: through
      the transform operator),
   2. A scaling function (directly from the transform's operator),

   and compute products to add them finally to appropriate alpha coefficients. 
...*/
            aplha[n - l] = y_k * phi(x_k, l);   // for l = 0,..., 2*p - 1
/**...
   It is clear that transform operator works here like compositor (cf.
   [BS, Section 22.4.7]) and threfore the main role in computations is played
   by the iterator's assigment operator:
...*/
      double operator = (pair<double, double> const &,
                         scaling_function<double, p, sf_traits<p> > const &);
/**...
   Voila (there it is!)!

   --
   ISSUES:
   
  */
